﻿using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;
using HeuristicLab.Encodings.MoveVectorEncoding.Interfaces;
using HeuristicLab.Data.MoveVectorData;

namespace HeuristicLab.Encodings.IntegerVectorEncoding
{
    /// <summary>
    /// A shaking operator for VNS.
    /// </summary>
    [Item("MoveVectorShakingOperator", "A shaking operator for VNS which uses available manipulation operators to perform the shaking.")]
    [StorableClass]
    public class MoveVectorShakingOperator : ShakingOperator<IMoveVectorManipulator>, IMoveVectorMultiNeighborhoodShakingOperator, IStochasticOperator, IBoundedMoveVectorOperator
    {

        public ILookupParameter<MoveVector> MoveVectorParameter
        {
            get { return (ILookupParameter<MoveVector>)Parameters["MoveVector"]; }
        }

        public ILookupParameter<IRandom> RandomParameter
        {
            get { return (LookupParameter<IRandom>)Parameters["Random"]; }
        }

        public IValueLookupParameter<IntMatrix> BoundsParameter
        {
            get { return (IValueLookupParameter<IntMatrix>)Parameters["Bounds"]; }
        }

        [StorableConstructor]
        protected MoveVectorShakingOperator(bool deserializing) : base(deserializing) { }
        protected MoveVectorShakingOperator(MoveVectorShakingOperator original, Cloner cloner) : base(original, cloner) { }
        public override IDeepCloneable Clone(Cloner cloner)
        {
            return new MoveVectorShakingOperator(this, cloner);
        }
        public MoveVectorShakingOperator()
          : base()
        {
            Parameters.Add(new LookupParameter<MoveVector>("MoveVector", "The move vector to shake."));
            Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator that will be used for stochastic shaking operators."));
            Parameters.Add(new ValueLookupParameter<IntMatrix>("Bounds", "A 2 column matrix specifying the lower and upper bound for each dimension. If there are less rows than dimension the bounds vector is cycled."));

            foreach (IMoveVectorManipulator shaker in ApplicationManager.Manager.GetInstances<IMoveVectorManipulator>().OrderBy(x => x.Name))
                if (!(shaker is ISelfAdaptiveManipulator)) Operators.Add(shaker);
        }

        [StorableHook(HookType.AfterDeserialization)]
        private void AfterDeserialization()
        {
            #region Backwards compatible code, remove with 3.4
            if (!Parameters.ContainsKey("Bounds"))
                Parameters.Add(new ValueLookupParameter<IntMatrix>("Bounds", "A 2 column matrix specifying the lower and upper bound for each dimension. If there are less rows than dimension the bounds vector is cycled."));
            #endregion
        }

        #region Wiring of some parameters
        protected override void Operators_ItemsAdded(object sender, CollectionItemsChangedEventArgs<IndexedItem<IMoveVectorManipulator>> e)
        {
            base.Operators_ItemsAdded(sender, e);
            ParameterizeOperators(e.Items);
        }

        protected override void Operators_ItemsReplaced(object sender, CollectionItemsChangedEventArgs<IndexedItem<IMoveVectorManipulator>> e)
        {
            base.Operators_ItemsReplaced(sender, e);
            ParameterizeOperators(e.Items);
        }

        private void ParameterizeOperators(IEnumerable<IndexedItem<IMoveVectorManipulator>> items)
        {
            if (items.Any())
            {
                foreach (IStochasticOperator op in items.Select(x => x.Value).OfType<IStochasticOperator>())
                    op.RandomParameter.ActualName = RandomParameter.Name;
                foreach (IMoveVectorManipulator op in items.Select(x => x.Value).OfType<IMoveVectorManipulator>())
                    op.MoveVectorParameter.ActualName = MoveVectorParameter.Name;
                foreach (IBoundedMoveVectorOperator op in items.Select(x => x.Value).OfType<IBoundedMoveVectorOperator>())
                    op.BoundsParameter.ActualName = "MoveVector." + BoundsParameter.Name;
            }
        }
        #endregion
    }
}